@@ -1,5 +1,7 @@
 import logging
 
+from celery import chain
+
 from django.contrib import admin
 from django.db import models
 from django.db.models.functions import Lower
@@ -18,6 +20,7 @@ from .models import (
     EveSolarSystem,
     EveMoon,
     EvePlanet,
+    EveSovereigntyMap,
     StructureTag,
     StructureService,
     Webhook,
@@ -76,6 +79,15 @@ if STRUCTURES_DEVELOPER_MODE:
     class EveTypeAdmin(admin.ModelAdmin):
         pass
 
+    @admin.register(EveSovereigntyMap)
+    class EveSovereigntyMapAdmin(admin.ModelAdmin):
+        list_display = (
+            'solar_system_id', 'alliance_id', 'corporation_id', 'faction_id'
+        )
+        search_fields = [
+            'solar_system_id', 'alliance_id', 'corporation_id', 'faction_id'
+        ]
+
 
 @admin.register(Notification)
 class NotificationAdmin(admin.ModelAdmin):
@@ -354,14 +366,16 @@ class OwnerAdmin(admin.ModelAdmin):
         "Fetch notifications from EVE server"
 
     def send_notifications(self, request, queryset):
-        for obj in queryset:
-            tasks.send_new_notifications_for_owner.delay(
-                obj.pk
-            )
+        send_tasks = list()
+        for owner in queryset:            
+            send_tasks.append(tasks.send_new_notifications_for_owner.si(
+                owner_pk=owner.pk
+            ))
             self.message_user(
                 request,
-                'Started sending new notifications for: {}. '.format(obj)
+                'Started sending new notifications for: {}. '.format(owner)
             )
+        chain(send_tasks).delay()
 
     send_notifications.short_description = "Send new notifications to Discord"
 
@@ -373,16 +387,25 @@ class OwnerAdmin(admin.ModelAdmin):
 class StructureTagAdmin(admin.ModelAdmin):
     list_display = (
         'name',
-        'description',
+        'description',        
+        'order',        
         'style',
         'is_default',
+        'is_user_managed',
     )
-    list_filter = (
-        'is_default',
+    list_filter = (        
+        'is_default',        
         'style',
-    )
-    ordering = ['name']
+        'is_user_managed',
+    )    
+    readonly_fields = ('is_user_managed',)
 
+    def has_delete_permission(self, request, obj=None):
+        return False if obj and not obj.is_user_managed else True
+
+    def has_change_permission(self, request, obj=None):
+        return False if obj and not obj.is_user_managed else True
+            
 
 class StructureAdminInline(admin.TabularInline):
     model = StructureService
@@ -476,37 +499,35 @@ class StructureAdmin(admin.ModelAdmin):
         ('tags', admin.RelatedOnlyFieldListFilter),
     )
 
-    actions = ('add_default_tags', 'remove_all_tags', )
+    actions = ('add_default_tags', 'remove_user_tags', 'update_generated_tags')
 
-    def _owner(self, obj):
+    def _owner(self, structure):
         return format_html(
             '{}<br>{}',
-            obj.owner.corporation,
-            obj.owner.corporation.alliance
+            structure.owner.corporation,
+            structure.owner.corporation.alliance
         )
 
-    def _location(self, obj):
-        if obj.eve_moon:
-            location_name = obj.eve_moon.name
-        elif obj.eve_planet:
-            location_name = obj.eve_planet.name
+    def _location(self, structure):
+        if structure.eve_moon:
+            location_name = structure.eve_moon.name
+        elif structure.eve_planet:
+            location_name = structure.eve_planet.name
         else:
-            location_name = obj.eve_solar_system.name
+            location_name = structure.eve_solar_system.name
         return format_html(
             '{}<br>{}',
             location_name,
-            obj.eve_solar_system.eve_constellation.eve_region
+            structure.eve_solar_system.eve_constellation.eve_region
         )
 
-    def _type(self, obj):
+    def _type(self, structure):
         return format_html(
-            '{}<br>{}',
-            obj.eve_type,
-            obj.eve_type.eve_group
+            '{}<br>{}', structure.eve_type, structure.eve_type.eve_group
         )
 
-    def _tags(self, obj):
-        tag_names = [x.name for x in obj.tags.all().order_by('name')]
+    def _tags(self, structure):
+        tag_names = [x.name for x in structure.tags.all()]
         if tag_names:
             return tag_names
         else:
@@ -520,9 +541,9 @@ class StructureAdmin(admin.ModelAdmin):
     def add_default_tags(self, request, queryset):
         structure_count = 0
         tags = StructureTag.objects.filter(is_default=True)
-        for obj in queryset:
+        for structure in queryset:
             for tag in tags:
-                obj.tags.add(tag)
+                structure.tags.add(tag)
             structure_count += 1
 
         self.message_user(
@@ -535,18 +556,32 @@ class StructureAdmin(admin.ModelAdmin):
     add_default_tags.short_description = \
         "Add default tags to selected structures"
 
-    def remove_all_tags(self, request, queryset):
+    def remove_user_tags(self, request, queryset):
         structure_count = 0
-        for obj in queryset:
-            obj.tags.clear()
+        for structure in queryset:
+            for tag in structure.tags.filter(is_user_managed=True):
+                structure.tags.remove(tag)
             structure_count += 1
 
         self.message_user(
             request,
-            'Removed all tags from {:,} structures'.format(structure_count))
+            'Removed all user tags from {:,} structures'.format(structure_count))
 
-    remove_all_tags.short_description = \
-        "Remove all tags from selected structures"
+    remove_user_tags.short_description = \
+        "Remove user tags for selected structures"
+
+    def update_generated_tags(self, request, queryset):
+        structure_count = 0
+        for structure in queryset:
+            structure.update_generated_tags(recreate_tags=True)
+            structure_count += 1
+
+        self.message_user(
+            request,
+            'Updated all generated tags for {:,} structures'.format(structure_count))
+
+    update_generated_tags.short_description = \
+        "Update generated tags for selected structures"
 
     if not STRUCTURES_DEVELOPER_MODE:
         readonly_fields = tuple([
@@ -593,6 +628,15 @@ class StructureAdmin(admin.ModelAdmin):
     )
     inlines = (StructureAdminInline, )
 
+    def formfield_for_manytomany(self, db_field, request, **kwargs):
+        """only show custom tags in dropdown"""
+        if db_field.name == "tags":
+            kwargs["queryset"] = StructureTag.objects.filter(is_user_managed=True)
+        
+        return super(StructureAdmin, self).formfield_for_manytomany(
+            db_field, request, **kwargs
+        )
+
 
 @admin.register(Webhook)
 class WebhookAdmin(admin.ModelAdmin):
