@@ -3,10 +3,13 @@ from pydoc import locate
 
 from bravado.exception import HTTPError
 
-from django.db import models
+from django.db import models, transaction
+from django.utils.timezone import now
+
+from esi.models import Token
 
 from . import __title__
-from .helpers import EsiSmartRequest
+from structures.helpers.esi_fetch import esi_fetch_with_localization, esi_fetch
 from .utils import LoggerAddTag, make_logger_prefix
 
 
@@ -48,16 +51,16 @@ class EveUniverseManager(models.Manager):
             args = {self.model.esi_pk(): eve_id}
             if self.model.has_esi_localization():
                 eve_data_objects = \
-                    EsiSmartRequest.fetch_with_localization(
+                    esi_fetch_with_localization(
                         esi_path=esi_path,
-                        args=args,                        
                         languages=EsiNameLocalization.ESI_LANGUAGES,
+                        args=args,                        
                         logger_tag=add_prefix()
                     )
             else:
                 eve_data_objects = dict()
                 eve_data_objects[EsiNameLocalization.ESI_DEFAULT_LANGUAGE] = \
-                    EsiSmartRequest.fetch(
+                    esi_fetch(
                         esi_path=esi_path,
                         args=args,
                         logger_tag=add_prefix()
@@ -100,6 +103,38 @@ class EveUniverseManager(models.Manager):
         return count_updated
 
 
+class EveSovereigntyMapManager(models.Manager):
+    
+    def update_from_esi(self):        
+        logger.info('Fetching sovereignty map from ESI...')
+        map = esi_fetch('Sovereignty.get_sovereignty_map', args={})
+        last_updated = now()
+        obj_list = list()
+        for solar_system in map:
+            obj_def = {
+                'solar_system_id': solar_system['system_id'],
+                'last_updated': last_updated
+            }
+            for key in ['alliance_id', 'corporation_id', 'faction_id']:
+                if key in solar_system and solar_system[key]:
+                    obj_def[key] = solar_system[key]
+                else:
+                    obj_def[key] = None
+
+            if (
+                obj_def['alliance_id']
+                or obj_def['corporation_id']
+                or obj_def['faction_id']
+            ):                
+                obj_list.append(self.model(**obj_def))
+        
+        if obj_list:
+            logger.info('Storing sovereignty map ...')
+            with transaction.atomic():
+                self.all().delete()
+                self.bulk_create(obj_list, batch_size=1000)
+                    
+
 class EveEntityManager(models.Manager):
 
     def get_or_create_esi(self, eve_entity_id: int) -> tuple:
@@ -108,12 +143,11 @@ class EveEntityManager(models.Manager):
         eve_id: Eve Online ID of object
         
         Returns: object, created        
-        """
-        from .models import EveEntity
+        """        
         try:
             obj = self.get(id=eve_entity_id)
             created = False
-        except EveEntity.DoesNotExist:
+        except self.model.DoesNotExist:
             obj, created = self.update_or_create_esi(eve_entity_id)
 
         return obj, created
@@ -129,7 +163,7 @@ class EveEntityManager(models.Manager):
             '%s(id=%d)' % (self.model.__name__, eve_entity_id)
         )        
         try:            
-            response = EsiSmartRequest.fetch(
+            response = esi_fetch(
                 esi_path='Universe.post_universe_names', 
                 args={'ids': [eve_entity_id]},
                 logger_tag=add_prefix()
@@ -159,34 +193,34 @@ class EveEntityManager(models.Manager):
 class StructureManager(models.Manager):
 
     def get_or_create_esi(
-        self, structure_id: int, esi_client: object
+        self, structure_id: int, token: Token
     ) -> tuple:
         """get or create a structure with data from ESI if needed
         
         structure_id: Structure ID of object in Eve Online
 
-        esi_client: ESI client with scope: esi-universe.read_structures.v1
+        token: ``esi.models.Token`` object with scope: 
+        ``esi-universe.read_structures.v1``
         
         Returns: object, created
-        """
-        from .models import Structure
-
+        """        
         try:
-            obj = Structure.objects.get(id=structure_id)
+            obj = self.get(id=structure_id)
             created = False
-        except Structure.DoesNotExist:            
-            obj, created = self.update_or_create_esi(structure_id, esi_client)
+        except self.model.DoesNotExist:            
+            obj, created = self.update_or_create_esi(structure_id, token)
         return obj, created
 
     def update_or_create_esi(
-        self, structure_id: int, esi_client: object
+        self, structure_id: int, token: Token
     ) -> tuple:
         """update or create a structure from ESI for given structure ID
         This will only fetch basic info about a structure
         
         structure_id: Structure ID of object in Eve Online
 
-        esi_client: ESI client with scope: esi-universe.read_structures.v1
+        token: ``esi.models.Token`` object with scope: 
+        ``esi-universe.read_structures.v1``
         
         Returns: object, created
         """
@@ -198,13 +232,13 @@ class StructureManager(models.Manager):
         logger.info(add_prefix('Trying to fetch structure from ESI'))
         
         try:
-            if esi_client is None:
-                raise ValueError('Can not fetch structure without esi client')
+            if token is None:
+                raise ValueError('Can not fetch structure without token')
             
-            structure_info = EsiSmartRequest.fetch(
+            structure_info = esi_fetch(
                 esi_path='Universe.get_universe_structures_structure_id', 
                 args={'structure_id': structure_id},                 
-                esi_client=esi_client,
+                token=token,
                 logger_tag=add_prefix()
             )            
             structure = {
@@ -233,16 +267,17 @@ class StructureManager(models.Manager):
         self, structure: dict, owner: object
     ) -> tuple:
         """update or create structure from given dict"""
-        from .models import EveType, EveSolarSystem, Structure,\
-            StructureService, EvePlanet, EveMoon
-        from .models.eveuniverse import EveUniverse
-        eve_type, _ = EveType.objects.get_or_create_esi(
-            structure['type_id']
+        from .models import (
+            EveType, 
+            EveSolarSystem, 
+            StructureService, 
+            EvePlanet, 
+            EveMoon
         )
+        from .models.eveuniverse import EveUniverse
+        eve_type, _ = EveType.objects.get_or_create_esi(structure['type_id'])
         eve_solar_system, _ = \
-            EveSolarSystem.objects.get_or_create_esi(
-                structure['system_id']
-            )
+            EveSolarSystem.objects.get_or_create_esi(structure['system_id'])
         fuel_expires = \
             structure['fuel_expires'] \
             if 'fuel_expires' in structure else None
@@ -264,8 +299,8 @@ class StructureManager(models.Manager):
             if 'reinforce_hour' in structure else None
 
         state = \
-            Structure.get_matching_state_for_esi_state(structure['state']) \
-            if 'state' in structure else Structure.STATE_UNKNOWN
+            self.model.get_matching_state_for_esi_state(structure['state']) \
+            if 'state' in structure else self.model.STATE_UNKNOWN
 
         state_timer_start = \
             structure['state_timer_start'] \
@@ -292,16 +327,14 @@ class StructureManager(models.Manager):
             if 'position' in structure else None
 
         if 'planet_id' in structure:
-            eve_planet, _ = EvePlanet.objects.get_or_create_esi(
-                structure['planet_id'],                
-            )
+            eve_planet, _ = \
+                EvePlanet.objects.get_or_create_esi(structure['planet_id'])
         else:
             eve_planet = None
 
         if 'moon_id' in structure:
-            eve_moon, _ = EveMoon.objects.get_or_create_esi(
-                structure['moon_id'],             
-            )
+            eve_moon, _ = \
+                EveMoon.objects.get_or_create_esi(structure['moon_id'])
         else:
             eve_moon = None
 
@@ -329,6 +362,7 @@ class StructureManager(models.Manager):
                 'last_updated': owner.structures_last_sync
             }
         )
+        # save related structure services
         StructureService.objects.filter(structure=obj).delete()
         if 'services' in structure and structure['services']:
             for service in structure['services']:
@@ -347,4 +381,73 @@ class StructureManager(models.Manager):
                             args[field_name] = service[field_name]
 
                 StructureService.objects.create(**args)
+                
+        return obj, created
+
+
+class StructureTagManager(models.Manager):
+    
+    def get_or_create_for_space_type(self, solar_system: object) -> tuple:
+        if solar_system.space_type in self.model.SPACE_TYPE_MAP:
+            name = self.model.SPACE_TYPE_MAP[solar_system.space_type]['name']
+        else:
+            name = None
+
+        if name:
+            try:
+                obj = self.get(name=name)
+                created = False
+            except self.model.DoesNotExist:
+                obj, created = self.update_or_create_for_space_type(solar_system)
+
+        else:
+            obj = None
+            created = None
+        
+        return obj, created
+
+    def update_or_create_for_space_type(self, solar_system: object) -> tuple:
+        if solar_system.space_type in self.model.SPACE_TYPE_MAP:
+            params = self.model.SPACE_TYPE_MAP[solar_system.space_type]            
+        else:
+            params = None
+        
+        if params:
+            return self.update_or_create(
+                name=params['name'], 
+                defaults={
+                    'style': params['style'],
+                    'description': (
+                        'this tag represents a space type. system generated.'
+                    ),                
+                    'order': 50,
+                    'is_user_managed': False,
+                    'is_default': False
+                }
+            )
+        else:
+            return None, None
+
+    def get_or_create_for_sov(self) -> tuple:
+        try:
+            obj = self.get(name=self.model.NAME_SOV_TAG)
+            created = False
+        except self.model.DoesNotExist:
+            obj, created = self.update_or_create_for_sov()
+
+        return obj, created
+    
+    def update_or_create_for_sov(self) -> tuple:
+        obj, created = self.update_or_create(
+            name=self.model.NAME_SOV_TAG, 
+            defaults={
+                'style': self.model.STYLE_DARK_BLUE,
+                'description': (
+                    'Owner of this structure has sovereignty. system generated.'
+                ),
+                'order': 20,
+                'is_user_managed': False,
+                'is_default': False
+            }
+        )
         return obj, created
