@@ -3,10 +3,13 @@
 from datetime import datetime, timedelta
 import json
 import logging
+import math
 import os
 import re
 from time import sleep
 
+from bravado.exception import HTTPError
+
 from django.core.serializers.json import DjangoJSONEncoder
 from django.conf import settings
 from django.contrib.auth.models import User
@@ -18,7 +21,6 @@ from allianceauth.authentication.models import CharacterOwnership
 from allianceauth.eveonline.models import EveCorporationInfo
 from allianceauth.notifications import notify
 
-from esi.clients import esi_client_factory
 from esi.errors import TokenExpiredError, TokenInvalidError, TokenError
 from esi.models import Token
 
@@ -36,7 +38,8 @@ from ..app_settings import (
     STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES,    
 )
 from .eveuniverse import EvePlanet, EveSolarSystem, EveType, EveUniverse
-from ..helpers import EsiSmartRequest
+from structures.helpers.esi_fetch import esi_fetch
+from structures.helpers.esi_fetch import esi_fetch_with_localization
 from .structures import Structure
 from .notifications import EveEntity, Notification
                 
@@ -44,8 +47,7 @@ from ..utils import (
     LoggerAddTag, 
     make_logger_prefix, 
     chunks, 
-    DATETIME_FORMAT, 
-    get_swagger_spec_path
+    DATETIME_FORMAT
 )
 
 logger = LoggerAddTag(logging.getLogger(__name__), __title__)
@@ -267,7 +269,7 @@ class Owner(models.Model):
         
         add_prefix = self._logger_prefix()
         try:            
-            esi_client, error = self.esi_client()
+            token, error = self.token()
             if error:
                 self.structures_last_error = error
                 self.structures_last_sync = now()
@@ -275,11 +277,11 @@ class Owner(models.Model):
                 raise TokenError(self.to_friendly_error_message(error))
 
             try:                
-                structures = self._fetch_upwell_structures(esi_client)
+                structures = self._fetch_upwell_structures(token)
                 if STRUCTURES_FEATURE_CUSTOMS_OFFICES:
-                    structures += self._fetch_custom_offices(esi_client)
+                    structures += self._fetch_custom_offices(token)
                 if STRUCTURES_FEATURE_STARBASES:
-                    structures += self._fetch_starbases(esi_client)
+                    structures += self._fetch_starbases(token)
 
                 logger.info(add_prefix(
                     'Storing updates for {:,} structures'.format(
@@ -341,7 +343,7 @@ class Owner(models.Model):
 
         return success
 
-    def _fetch_upwell_structures(self, esi_client: object) -> list:
+    def _fetch_upwell_structures(self, token: Token) -> list:
         """fetch Upwell structures from ESI for self"""
         from .eveuniverse import EsiNameLocalization
 
@@ -349,12 +351,12 @@ class Owner(models.Model):
         corporation_id = self.corporation.corporation_id
         
         # fetch all structures incl. localizations for services
-        structures_w_lang = EsiSmartRequest.fetch_with_localization(
+        structures_w_lang = esi_fetch_with_localization(
             esi_path='Corporation.get_corporations_corporation_id_structures',
-            args={'corporation_id': corporation_id},            
+            args={'corporation_id': corporation_id}, 
+            token=token,
             languages=EsiNameLocalization.ESI_LANGUAGES,
-            has_pages=True,            
-            esi_client=esi_client,
+            has_pages=True,
             logger_tag=add_prefix()
         )
         
@@ -376,16 +378,25 @@ class Owner(models.Model):
                 )
             ))
             for structure in structures:                
-                structure_info = EsiSmartRequest.fetch(
-                    'Universe.get_universe_structures_structure_id',
-                    args={'structure_id': structure['structure_id']},                    
-                    esi_client=esi_client,
-                    logger_tag=add_prefix()
-                )                
-                structure['name'] = Structure.extract_name_from_esi_respose(
-                    structure_info['name']
-                )
-                structure['position'] = structure_info['position']
+                try:
+                    structure_info = esi_fetch(
+                        'Universe.get_universe_structures_structure_id',
+                        args={'structure_id': structure['structure_id']},
+                        token=token,
+                        logger_tag=add_prefix()
+                    )
+                    structure['name'] = Structure.extract_name_from_esi_respose(
+                        structure_info['name']
+                    )
+                    structure['position'] = structure_info['position']
+                except HTTPError as ex:
+                    logger.exception(
+                        'Failed to load details for structure with ID %d due '
+                        'to the following exception: %s' % (
+                            structure['structure_id'], ex
+                        )
+                    )
+                    structure['name'] = '(no data)'
 
         if STRUCTURES_DEVELOPER_MODE:
             self._store_raw_data('structures', structures, corporation_id)
@@ -447,23 +458,16 @@ class Owner(models.Model):
                             service['name_' + lang] = name_loc
         return structures
 
-    def _fetch_custom_offices(self, esi_client: object) -> list:
-        """fetch custom offices from ESI for self"""
-
-        def extract_planet_name(text: str) -> str:
-            """extract name of planet from assert name for a customs office"""
-            reg_ex = re.compile(r'Customs Office \((.+)\)')
-            matches = reg_ex.match(text)
-            return matches.group(1) if matches else text
-        
+    def _fetch_custom_offices(self, token: Token) -> list:
+        """fetch custom offices from ESI for self"""        
         add_prefix = self._logger_prefix()        
         corporation_id = self.corporation.corporation_id
 
-        pocos = EsiSmartRequest.fetch(
+        pocos = esi_fetch(
             'Planetary_Interaction.get_corporations_corporation_id_customs_offices',
-            args={'corporation_id': corporation_id},            
+            args={'corporation_id': corporation_id},
+            token=token,
             has_pages=True,
-            esi_client=esi_client,
             logger_tag=add_prefix()
         )       
         structures = list()
@@ -472,48 +476,13 @@ class Owner(models.Model):
                 'No custom offices retrieved from ESI'
             ))
         else:
-            # fetching locations
-            logger.info(add_prefix(
-                'Fetching locations for {} custom offices from ESI'.format(
-                    len(pocos)
-                )
-            ))
-            item_ids = [x['office_id'] for x in pocos]
-            locations_data = list()
-            for item_ids_chunk in chunks(item_ids, 999):               
-                locations_data_chunk = EsiSmartRequest.fetch(
-                    'Assets.post_corporations_corporation_id_assets_locations',
-                    args={
-                        'corporation_id': corporation_id, 
-                        'item_ids': item_ids_chunk
-                    },                    
-                    esi_client=esi_client,
-                    logger_tag=add_prefix()
-                )
-                locations_data += locations_data_chunk
-            positions = {x['item_id']: x['position'] for x in locations_data}
-
-            # fetching names
-            logger.info(add_prefix(
-                'Fetching names for {} custom office names from ESI'.format(
-                    len(pocos)
-                )
-            ))
-            names_data = list()
-            for item_ids_chunk in chunks(item_ids, 999):                
-                names_data_chunk = EsiSmartRequest.fetch(
-                    'Assets.post_corporations_corporation_id_assets_names',
-                    args={
-                        'corporation_id': corporation_id, 
-                        'item_ids': item_ids
-                    },                    
-                    esi_client=esi_client,
-                    logger_tag=add_prefix()
-                )
-                names_data += names_data_chunk
-            names = {
-                x['item_id']: extract_planet_name(x['name']) for x in names_data
-            }
+            item_ids = [x['office_id'] for x in pocos]            
+            positions = self._fetch_locations_for_pocos(
+                corporation_id, item_ids, token
+            )
+            names = self._fetch_names_for_pocos(
+                corporation_id, item_ids, token
+            )
 
             # making sure we have all solar systems loaded
             # incl. their planets for later name matching
@@ -567,40 +536,86 @@ class Owner(models.Model):
 
         return structures
 
-    def _fetch_starbases(self, esi_client: object) -> list:
+    def _fetch_locations_for_pocos(self, corporation_id, item_ids, token):
+        add_prefix = self._logger_prefix()
+        logger.info(add_prefix(
+            'Fetching locations for {} custom offices from ESI'.format(
+                len(item_ids)
+            )
+        ))        
+        locations_data = list()
+        for item_ids_chunk in chunks(item_ids, 999):               
+            locations_data_chunk = esi_fetch(
+                'Assets.post_corporations_corporation_id_assets_locations',
+                args={
+                    'corporation_id': corporation_id, 
+                    'item_ids': item_ids_chunk,
+                },
+                token=token,
+                logger_tag=add_prefix()
+            )
+            locations_data += locations_data_chunk
+        positions = {x['item_id']: x['position'] for x in locations_data}
+        return positions
+
+    def _fetch_names_for_pocos(self, corporation_id, item_ids, token):
+        add_prefix = self._logger_prefix()
+        logger.info(add_prefix(
+            'Fetching names for {} custom office names from ESI'.format(
+                len(item_ids)
+            )
+        ))
+        names_data = list()
+        for item_ids_chunk in chunks(item_ids, 999):                
+            names_data_chunk = esi_fetch(
+                'Assets.post_corporations_corporation_id_assets_names',
+                args={
+                    'corporation_id': corporation_id, 
+                    'item_ids': item_ids,                        
+                },
+                token=token,
+                logger_tag=add_prefix()
+            )
+            names_data += names_data_chunk
+
+        names = {
+            x['item_id']: self._extract_planet_name(x['name']) 
+            for x in names_data
+        }
+        return names
+
+    @staticmethod
+    def _extract_planet_name(text: str) -> str:
+        """extract name of planet from assert name for a customs office"""
+        reg_ex = re.compile(r'Customs Office \((.+)\)')
+        matches = reg_ex.match(text)
+        return matches.group(1) if matches else text
+
+    def _fetch_starbases(self, token: Token) -> list:
         """fetch starbases from ESI for self"""
 
         add_prefix = self._logger_prefix()        
-        corporation_id = self.corporation.corporation_id
-        starbases = EsiSmartRequest.fetch(
+        structures = list()
+        corporation_id = self.corporation.corporation_id        
+        starbases = esi_fetch(
             'Corporation.get_corporations_corporation_id_starbases',
-            args={'corporation_id': corporation_id},            
+            args={'corporation_id': corporation_id},
+            token=token,
             has_pages=True,
-            esi_client=esi_client,
             logger_tag=add_prefix()
         )
-        # convert into structures data format
-        structures = list()
+               
         if not starbases:
             logger.info(add_prefix('No starbases retrieved from ESI'))
         else:
-            logger.info(add_prefix(
-                'Fetching names for {} starbases from ESI'.format(len(starbases))
-            ))
-            item_ids = [x['starbase_id'] for x in starbases]
-            names_data = list()
-            for item_ids_chunk in chunks(item_ids, 999):
-                names_data_chunk = EsiSmartRequest.fetch(
-                    'Assets.post_corporations_corporation_id_assets_names',
-                    args={
-                        'corporation_id': corporation_id, 
-                        'item_ids': item_ids_chunk
-                    },                    
-                    esi_client=esi_client,
-                    logger_tag=add_prefix()
+            names = self._fetch_starbases_names(
+                corporation_id, starbases, token
+            )            
+            for starbase in starbases:
+                starbase['fuel_expires'] = self._calc_starbase_fuel_expires(
+                    corporation_id, starbase, token
                 )
-                names_data += names_data_chunk
-            names = {x['item_id']: x['name'] for x in names_data}
+            # convert starbases to structures            
             for starbase in starbases:
                 if starbase['starbase_id'] in names:
                     name = names[starbase['starbase_id']]
@@ -619,6 +634,9 @@ class Owner(models.Model):
                 if 'moon_id' in starbase:
                     structure['moon_id'] = starbase['moon_id']
 
+                if 'fuel_expires' in starbase:
+                    structure['fuel_expires'] = starbase['fuel_expires']
+
                 if 'reinforced_until' in starbase:
                     structure['state_timer_end'] = starbase['reinforced_until']
 
@@ -631,6 +649,75 @@ class Owner(models.Model):
             self._store_raw_data('starbases', structures, corporation_id)
 
         return structures
+
+    def _fetch_starbases_names(self, corporation_id, starbases, token):
+        add_prefix = self._logger_prefix()
+        logger.info(add_prefix(
+            'Fetching names for {} starbases from ESI'.format(len(starbases))
+        ))
+        item_ids = [x['starbase_id'] for x in starbases]
+        names_data = list()
+        for item_ids_chunk in chunks(item_ids, 999):
+            names_data_chunk = esi_fetch(
+                'Assets.post_corporations_corporation_id_assets_names',
+                args={
+                    'corporation_id': corporation_id, 
+                    'item_ids': item_ids_chunk,                        
+                },
+                token=token,
+                logger_tag=add_prefix()
+            )
+            names_data += names_data_chunk
+        names = {x['item_id']: x['name'] for x in names_data}
+
+        for starbase in starbases:                
+            starbase_id = starbase['starbase_id']
+            starbase['name'] = \
+                names[starbase_id] if starbase_id in names else 'Starbase'
+
+        return names
+
+    def _calc_starbase_fuel_expires(self, corporation_id, starbase, token):
+        add_prefix = self._logger_prefix()
+        fuel_expires = None
+        if starbase['state'] != 'offline':
+            starbase_details = esi_fetch(
+                'Corporation.get_corporations_corporation_id_starbases_starbase_id',
+                args={
+                    'corporation_id': corporation_id, 
+                    'starbase_id': starbase['starbase_id'],
+                    'system_id': starbase['system_id'],
+                },
+                token=token,
+                logger_tag=add_prefix()
+            )
+            fuel_quantity = None
+            if 'fuels' in starbase_details:
+                for fuel in starbase_details['fuels']:
+                    fuel_type, _ = EveType.objects.get_or_create_esi(
+                        fuel['type_id']
+                    )
+                    if fuel_type.is_fuel_block:
+                        fuel_quantity = fuel['quantity']
+            if fuel_quantity:                    
+                starbase_type, _ = EveType.objects.get_or_create_esi(
+                    starbase['type_id']
+                )
+                solar_system, _ = \
+                    EveSolarSystem.objects.get_or_create_esi(
+                        starbase['system_id']
+                    )                                        
+                sov_discount = 0.25 \
+                    if solar_system.corporation_has_sov(self.corporation) else 0
+                hours = math.floor(
+                    fuel_quantity / (
+                        starbase_type.starbase_fuel_per_hour 
+                        * (1 - sov_discount)
+                    )
+                )
+                fuel_expires = now() + timedelta(hours=hours)
+        
+        return fuel_expires
     
     def fetch_notifications_esi(self, user: User = None):
         """fetches notification for the current owners and proceses them"""
@@ -638,7 +725,7 @@ class Owner(models.Model):
         add_prefix = self._logger_prefix()
         notifications_count = 0
         try:            
-            esi_client, error = self.esi_client()
+            token, error = self.token()
             if error:
                 self.notifications_last_error = error
                 self.notifications_last_sync = now()
@@ -647,7 +734,7 @@ class Owner(models.Model):
            
             # fetch notifications from ESI
             try:
-                notifications = self._fetch_notifications_from_esi(esi_client)
+                notifications = self._fetch_notifications_from_esi(token)
                 new_notifications_count, notifications_count = \
                     self._store_notifications(
                         notifications, notifications_count
@@ -658,7 +745,7 @@ class Owner(models.Model):
                             new_notifications_count
                         )
                     ))
-                    self._process_timers_for_notifications(esi_client)
+                    self._process_timers_for_notifications(token)
 
                 else:
                     logger.info(add_prefix(
@@ -696,15 +783,13 @@ class Owner(models.Model):
 
         return success
 
-    def _fetch_notifications_from_esi(self, esi_client):
+    def _fetch_notifications_from_esi(self, token: Token) -> dict:
         """ fetching all notifications from ESI for current owner"""
         add_prefix = self._logger_prefix()
-        notifications = EsiSmartRequest.fetch(
+        notifications = esi_fetch(
             'Character.get_characters_character_id_notifications',
-            args={
-                'character_id': self.character.character.character_id
-            },            
-            esi_client=esi_client,
+            args={'character_id': self.character.character.character_id},
+            token=token,
             logger_tag=add_prefix()
         )                
         if STRUCTURES_DEVELOPER_MODE:
@@ -799,7 +884,7 @@ class Owner(models.Model):
             
         return new_notifications_count, notifications_count
 
-    def _process_timers_for_notifications(self, esi_client):
+    def _process_timers_for_notifications(self, token: Token):
         """processes notifications for timers if any"""
         if STRUCTURES_ADD_TIMERS:
             cutoff_dt_for_stale = now() - timedelta(
@@ -814,11 +899,11 @@ class Owner(models.Model):
                 .select_related().order_by('timestamp')
 
             if len(notifications) > 0:
-                if not esi_client:
-                    esi_client = self.esi_client()
+                if not token:
+                    token = self.token()
 
                 for notification in notifications:
-                    notification.process_for_timerboard(esi_client)
+                    notification.process_for_timerboard(token)
 
     def send_new_notifications(
         self, rate_limited: bool = True, user: User = None
@@ -921,8 +1006,8 @@ class Owner(models.Model):
         """returns standard logger prefix function"""
         return make_logger_prefix(self.corporation.corporation_ticker)
 
-    def esi_client(self) -> object:
-        """returns a new ESI client for the given owner with token"""
+    def token(self) -> Token:
+        """returns a valid Token for the owner"""
         token = None
         error = None
         add_prefix = self._logger_prefix()
@@ -966,17 +1051,8 @@ class Owner(models.Model):
                         'No token found with sufficient scopes'
                     ))
                     error = self.ERROR_TOKEN_INVALID
-
-        if token:
-            logger.info(add_prefix('Starting ESI client...'))
-            logger.debug('Using token: {}'.format(token))                
-            client = esi_client_factory(
-                token=token, spec_file=get_swagger_spec_path()
-            )
-        else:
-            client = None
-        
-        return client, error
+                
+        return token, error
 
     def _send_report_to_user(
         self,        
