@@ -4,14 +4,19 @@ import re
 import logging
 
 from django.db import models
+from django.core.validators import MinValueValidator
 from django.core.validators import MaxValueValidator
-from django.utils.html import escape, format_html
+from django.utils.html import escape
 from django.utils.translation import gettext_lazy as _
+from django.utils.translation import gettext_noop
 
 from .. import __title__
-from ..managers import StructureManager
-from ..utils import LoggerAddTag
 from .eveuniverse import EsiNameLocalization
+from .eveuniverse import EveSolarSystem
+from ..managers import StructureManager
+from ..managers import StructureTagManager
+from ..utils import LoggerAddTag, create_bs_label_html
+
 
 logger = LoggerAddTag(logging.getLogger(__name__), __title__)
 
@@ -19,14 +24,47 @@ logger = LoggerAddTag(logging.getLogger(__name__), __title__)
 class StructureTag(models.Model):
     """tag for organizing structures"""
 
+    # special tags
+    NAME_SOV_TAG = gettext_noop('sov')
+    NAME_HIGHSEC_TAG = gettext_noop('highsec')
+    NAME_LOWSEC_TAG = gettext_noop('lowsec')
+    NAME_NULLSEC_TAG = gettext_noop('nullsec')
+    NAME_W_SPACE_TAG = gettext_noop('w_space')
+    
+    # styles
+    STYLE_GREY = 'default'
+    STYLE_DARK_BLUE = 'primary'
+    STYLE_GREEN = 'success'
+    STYLE_LIGHT_BLUE = 'info'
+    STYLE_ORANGE = 'warning'
+    STYLE_RED = 'danger'
     STYLE_CHOICES = [
-        ('default', 'grey'),
-        ('primary', 'dark blue'),
-        ('success', 'green'),
-        ('info', 'light blue'),
-        ('warning', 'yellow'),
-        ('danger', 'red'),
+        (STYLE_GREY, 'grey'),
+        (STYLE_DARK_BLUE, 'dark blue'),
+        (STYLE_GREEN, 'green'),
+        (STYLE_LIGHT_BLUE, 'light blue'),
+        (STYLE_ORANGE, 'orange'),
+        (STYLE_RED, 'red'),
     ]
+    
+    SPACE_TYPE_MAP = {
+        EveSolarSystem.TYPE_HIGHSEC: {
+            'name': NAME_HIGHSEC_TAG,
+            'style': STYLE_GREEN
+        },
+        EveSolarSystem.TYPE_LOWSEC: {
+            'name': NAME_LOWSEC_TAG,
+            'style': STYLE_ORANGE
+        },
+        EveSolarSystem.TYPE_NULLSEC: {
+            'name': NAME_NULLSEC_TAG,
+            'style': STYLE_RED
+        },
+        EveSolarSystem.TYPE_W_SPACE: {
+            'name': NAME_W_SPACE_TAG,
+            'style': STYLE_LIGHT_BLUE
+        },        
+    }
 
     name = models.CharField(
         max_length=255,
@@ -46,12 +84,31 @@ class StructureTag(models.Model):
         blank=True,
         help_text='color style of tag'
     )
+    order = models.PositiveIntegerField(
+        default=100,
+        blank=True,
+        validators=[MinValueValidator(100)],
+        help_text=(
+            'number defining the order tags are shown. '
+            'custom tags can not have an order below 100'
+        )
+    )    
     is_default = models.BooleanField(
         default=False,
         help_text=(
-            'if true this tag will automatically be added to new structures'
+            'if true this custom tag will automatically be added '
+            'to new structures'
         )
     )
+    is_user_managed = models.BooleanField(
+        default=True,
+        help_text=(
+            'if False this tag is created and managed by the system '
+            'and can not be modified by users'
+        )
+    )
+
+    objects = StructureTagManager()
 
     def __str__(self) -> str:
         return self.name
@@ -62,12 +119,16 @@ class StructureTag(models.Model):
             self.name
         )
 
+    class Meta:
+        ordering = ordering = ['order', 'name']
+
     @property
     def html(self) -> str:
-        return format_html('<span class="label label-{}">{}</span>'.format(
-            self.style,
-            escape(self.name)
-        ))
+        if self.is_user_managed:
+            name = escape(self.name)
+        else:
+            name = _(self.name)
+        return create_bs_label_html(name, self.style)
 
     @classmethod
     def sorted(cls, tags: list, reverse: bool = False) -> list:
@@ -298,7 +359,8 @@ class Structure(models.Model):
 
     @property
     def is_low_power(self):
-        return False if self.eve_type.is_poco else not self.fuel_expires
+        return False if not self.eve_type.is_upwell_structure \
+            else not self.fuel_expires
 
     @property
     def is_reinforced(self):
@@ -309,6 +371,10 @@ class Structure(models.Model):
             self.STATE_HULL_VULNERABLE
         ]
 
+    @property
+    def owner_has_sov(self):
+        return self.eve_solar_system.corporation_has_sov(self.owner.corporation)
+
     def __str__(self):
         return '{} - {}'.format(self.eve_solar_system, self.name)
 
@@ -333,6 +399,32 @@ class Structure(models.Model):
         """extracts the structure's name from the name in an ESI response"""
         matches = re.search(r'^\S+ - (.+)', esi_name)
         return matches.group(1) if matches else esi_name
+
+    def save(self, *args, **kwargs):
+        """make sure related objects are saved whenever structure is saved"""
+        super().save(*args, **kwargs)
+        self.update_generated_tags()
+        
+    def update_generated_tags(self, recreate_tags=False):
+        """updates all generated tags for this structure
+        
+        recreate_tags: when set true all tags will be re-created,
+        otherwise just re-added if they are missing
+        """        
+        method_name = 'update_or_create_for_space_type' \
+            if recreate_tags else 'get_or_create_for_space_type'
+
+        space_type_tag, _ = getattr(StructureTag.objects, method_name)(
+            self.eve_solar_system
+        )
+        
+        self.tags.add(space_type_tag)
+        if self.owner_has_sov:
+            method_name = 'update_or_create_for_sov' \
+                if recreate_tags else 'get_or_create_for_sov'
+
+            sov_tag, _ = getattr(StructureTag.objects, method_name)()
+            self.tags.add(sov_tag)
         
         
 class StructureService(EsiNameLocalization, models.Model):
