@@ -30,7 +30,8 @@ from .utils import (
     LoggerAddTag,
     timeuntil_str,
     add_no_wrap_html,
-    yesno_str
+    yesno_str,
+    create_bs_label_html
 )
 
 
@@ -80,7 +81,7 @@ def structure_list(request):
         if tags_raw:
             tags_parsed = tags_raw.split(',')
             active_tags = [
-                x for x in StructureTag.objects.all().order_by('name')
+                x for x in StructureTag.objects.all()
                 if x.name in tags_parsed
             ]
 
@@ -95,29 +96,28 @@ def structure_list(request):
     return render(request, 'structures/structure_list.html', context)
 
 
-class StructuresRowConverter:
-    """This class converts structure objects into rows for an HTML table"""
+class StructuresRowBuilder:
+    """This class build the HTML table rows from structure objects"""
     def __init__(self):
         self._row = None
         self._structure = None
     
     def convert(self, structure) -> dict:
         self._row = {
-            'structure_id': structure.id,
-            'is_poco': structure.eve_type.is_poco
+            'structure_id': structure.id
         }
         self._structure = structure
-        self._convert_owner()
-        self._convert_location()
-        self._convert_type()
-        self._convert_name()
-        self._convert_services()
-        self._convert_reinforcement_infos()
-        self._convert_fuel_infos()
-        self._convert_state()
+        self._build_owner()
+        self._build_location()
+        self._build_type()
+        self._build_name()
+        self._build_services()
+        self._build_reinforcement_infos()
+        self._build_fuel_infos()
+        self._build_state()
         return self._row
 
-    def _convert_owner(self):
+    def _build_owner(self):
         corporation = self._structure.owner.corporation
         if corporation.alliance:
             alliance_name = corporation.alliance.alliance_name
@@ -139,14 +139,17 @@ class StructuresRowConverter:
         self._row['alliance_name'] = alliance_name
         self._row['corporation_name'] = corporation.corporation_name
         
-    def _convert_location(self):
+    def _build_location(self):        
+        solar_system = self._structure.eve_solar_system
+        
+        # location
         self._row['region_name'] = \
-            self._structure.eve_solar_system.eve_constellation\
+            solar_system.eve_constellation\
             .eve_region.name_localized
         self._row['solar_system_name'] = \
-            self._structure.eve_solar_system.name_localized
+            solar_system.name_localized
         solar_system_url = dotlan.solar_system_url(
-            self._structure.eve_solar_system.name
+            solar_system.name
         )
         if self._structure.eve_moon:
             location_name = self._structure.eve_moon.name_localized
@@ -162,9 +165,10 @@ class StructuresRowConverter:
             add_no_wrap_html(self._row['region_name'])
         )
 
-    def _convert_type(self):
+    def _build_type(self):
+        structure_type = self._structure.eve_type
         # category
-        my_group = self._structure.eve_type.eve_group
+        my_group = structure_type.eve_group
         self._row['group_name'] = my_group.name_localized
         if my_group.eve_category:
             my_category = my_group.eve_category
@@ -177,7 +181,7 @@ class StructuresRowConverter:
         # type icon
         self._row['type_icon'] = format_html(
             '<img src="{}" width="{}" height="{}"/>',
-            self._structure.eve_type.icon_url(
+            structure_type.icon_url(
                 size=STRUCTURE_LIST_ICON_RENDER_SIZE
             ),
             STRUCTURE_LIST_ICON_OUTPUT_SIZE,
@@ -185,24 +189,28 @@ class StructuresRowConverter:
         )
 
         # type name
-        self._row['type_name'] = self._structure.eve_type.name_localized
+        self._row['type_name'] = structure_type.name_localized
         self._row['type'] = format_html(
             '{}<br>{}',
             add_no_wrap_html(self._row['type_name']),
             add_no_wrap_html(self._row['group_name'])
         )
 
-    def _convert_name(self):
-        self._row['structure_name'] = escape(self._structure.name)
+        # poco
+        self._row['is_poco'] = structure_type.is_poco
+
+    def _build_name(self):
+        self._row['structure_name'] = escape(self._structure.name)                
+        tags = []        
         if self._structure.tags:
+            tags += [
+                x.html for x in self._structure.tags.all()
+            ]
             self._row['structure_name'] += format_html(
-                '<br>{}',
-                mark_safe(' '.join([
-                    x.html for x in self._structure.tags.all().order_by('name')
-                ]))
+                '<br>{}', mark_safe(' '.join(tags))
             )
 
-    def _convert_services(self):
+    def _build_services(self):
         if self._row['is_poco'] or self._row['is_starbase']:
             self._row['services'] = gettext_lazy('N/A')
         else:
@@ -219,7 +227,7 @@ class StructuresRowConverter:
                 services.append(service_name)
             self._row['services'] = '<br>'.join(services)
 
-    def _convert_reinforcement_infos(self):
+    def _build_reinforcement_infos(self):
         self._row['is_reinforced'] = self._structure.is_reinforced
         self._row['is_reinforced_str'] = \
             yesno_str(self._structure.is_reinforced)
@@ -234,42 +242,37 @@ class StructuresRowConverter:
             else:
                 self._row['reinforcement'] = ''
 
-    def _convert_fuel_infos(self):
+    def _build_fuel_infos(self):
         self._row['is_low_power'] = self._structure.is_low_power
         self._row['is_low_power_str'] = yesno_str(self._structure.is_low_power)
 
-        if self._row['is_poco'] or self._row['is_starbase']:
+        if self._structure.is_low_power:
+            fuel_expires_display = format_html_lazy(
+                create_bs_label_html(gettext_lazy('Low Power'), 'default')
+            )                    
+            fuel_expires_timestamp = None
+        elif self._structure.fuel_expires:
+            fuel_expires_timestamp = self._structure.fuel_expires.isoformat()
+            if STRUCTURES_SHOW_FUEL_EXPIRES_RELATIVE:
+                fuel_expires_display = timeuntil_str(
+                    self._structure.fuel_expires - now()
+                )
+                if not fuel_expires_display:
+                    fuel_expires_display = '?'
+                    fuel_expires_timestamp = None
+            else:
+                fuel_expires_display = \
+                    self._structure.fuel_expires.strftime(DATETIME_FORMAT)
+        else:
             fuel_expires_display = gettext_lazy('N/A')
             fuel_expires_timestamp = None
-        else:
-            if self._row['is_low_power']:
-                fuel_expires_display = format_html_lazy(
-                    '<span class="label label-default">{}</span>',
-                    gettext_lazy('Low Power')
-                )                    
-                fuel_expires_timestamp = None
-            elif self._structure.fuel_expires:
-                fuel_expires_timestamp = self._structure.fuel_expires.isoformat()
-                if STRUCTURES_SHOW_FUEL_EXPIRES_RELATIVE:
-                    fuel_expires_display = timeuntil_str(
-                        self._structure.fuel_expires - now()
-                    )
-                    if not fuel_expires_display:
-                        fuel_expires_display = '?'
-                        fuel_expires_timestamp = None
-                else:
-                    fuel_expires_display = \
-                        self._structure.fuel_expires.strftime(DATETIME_FORMAT)
-            else:
-                fuel_expires_display = '?'
-                fuel_expires_timestamp = None
 
         self._row['fuel_expires'] = {
             'display': add_no_wrap_html(fuel_expires_display),
             'timestamp': fuel_expires_timestamp
         }
 
-    def _convert_state(self):
+    def _build_state(self):
         self._row['state_str'] = self._structure.get_state_display()
         self._row['state_details'] = self._row['state_str']
         if self._structure.state_timer_end:
@@ -287,7 +290,7 @@ def structure_list_data(request):
     """returns structure list in JSON for AJAX call in structure_list view"""
 
     structure_rows = list()
-    row_converter = StructuresRowConverter()
+    row_converter = StructuresRowBuilder()
     for structure in _structures_query_for_user(request):        
         structure_rows.append(row_converter.convert(structure))
 
